pub trait WeakFunctor {
    type F<A>;
}

pub trait Functor: WeakFunctor {
    fn fmap<A, B, F: FnOnce(A) -> B>(f: F, fa: Self::F<A>) -> Self::F<B>;

    fn replace<A, B>(fa: Self::F<A>, b: B) -> Self::F<B> {
        Self::fmap(|_| b, fa)
    }

    fn void<A>(fa: Self::F<A>) -> Self::F<()> {
        Self::replace(fa, ())
    }
}

pub trait ApplicativeSeq: Functor {
    fn seq<A, B, F: FnOnce(A) -> B>(ff: Self::F<F>, fa: Self::F<A>) -> Self::F<B>;
    fn _la2<A, B, C, F: FnOnce(A, B) -> C>(f: F, fa: Self::F<A>, fb: Self::F<B>) -> Self::F<C> {
        Self::seq(Self::fmap(|a| |b| f(a, b), fa), fb)
    }
}

pub trait ApplicativeLA2: Functor {
    fn la2<A, B, C, F: FnOnce(A, B) -> C>(f: F, fa: Self::F<A>, fb: Self::F<B>) -> Self::F<C>;
    fn _seq<A, B, F: FnOnce(A) -> B>(ff: Self::F<F>, fa: Self::F<A>) -> Self::F<B> {
        Self::la2(|f, a| f(a), ff, fa)
    }
}

pub trait Applicative: Functor + ApplicativeSeq + ApplicativeLA2 {
    fn pure<A>(a: A) -> Self::F<A>;

    fn discard_first<A, B>(fa: Self::F<A>, fb: Self::F<B>) -> Self::F<B> {
        Self::seq(Self::replace(fa, |b| b), fb)
    }

    fn discard_second<A, B>(fa: Self::F<A>, fb: Self::F<B>) -> Self::F<A> {
        Self::la2(|a, _| a, fa, fb)
    }
}

pub trait Monad: Applicative {
    fn bind<A, B, F: FnOnce(A) -> Self::F<B>>(fa: Self::F<A>, f: F) -> Self::F<B>;

    fn join<A>(ffa: Self::F<Self::F<A>>) -> Self::F<A> {
        Self::bind(ffa, |fa| fa)
    }
}

#[cfg(test)]
pub mod tests {
    use super::*;
    use core::fmt::Debug;
    use std::ops::{Add, AddAssign};

    pub struct TestResults(usize, usize, Result<(), String>);

    pub type R = TestResults;

    impl R {
        pub fn unwrap(self) {
            match self {
                TestResults(success, total, Err(e)) => panic!("failed:\n{success}/{total}\n{e}\n"),
                _ => (),
            };
        }
    }

    fn eqr<T: PartialEq + Debug>(name: &str, left: T, right: T) -> R {
        if left == right {
            TestResults(1, 1, Ok(()))
        } else {
            TestResults(0, 1, Err(format!("{name}: {:?} != {:?}", left, right)))
        }
    }

    impl Default for R {
        fn default() -> Self {
            Self(0, 0, Ok(()))
        }
    }

    impl Add for R {
        type Output = R;

        fn add(self, rhs: Self) -> Self::Output {
            Self(
                self.0 + rhs.0,
                self.1 + rhs.1,
                match (self.2, rhs.2) {
                    (Err(le), Err(re)) => Err(le + "\n" + re.as_str()),
                    (e, Ok(_)) => e,
                    (Ok(_), e) => e,
                },
            )
        }
    }

    impl AddAssign<R> for R {
        fn add_assign(&mut self, rhs: R) {
            self.0 += rhs.0;
            self.1 += rhs.1;
            match (&mut self.2, rhs.2) {
                (Err(ref mut le), Err(re)) => {
                    *le += "\n";
                    *le += re.as_str()
                }
                (_, Ok(_)) => {}
                (lr, Err(re)) => *lr = Err(re),
            }
        }
    }

    pub fn fmap_respects_identity<T: Functor, A, FA0: Fn() -> T::F<A>>(fa0: FA0) -> R
    where
        T::F<A>: Debug + PartialEq,
    {
        eqr("identity: fmap id == id", T::fmap(|a| a, fa0()), fa0())
    }

    pub fn fmap_respects_composition<
        T: Functor,
        A,
        B,
        C,
        F: Fn(B) -> C,
        G: Fn(A) -> B,
        FA0: Fn() -> T::F<A>,
    >(
        f: F,
        g: G,
        fa0: FA0,
    ) -> R
    where
        T::F<C>: Debug + PartialEq,
    {
        eqr(
            "composition: fmap (f . g) == fmap f . fmap g",
            T::fmap(|a| f(g(a)), fa0()),
            T::fmap(f, T::fmap(g, fa0())),
        )
    }

    pub fn seq_respects_identity<T: Applicative, A, FA0: Fn() -> T::F<A>>(fa0: FA0) -> R
    where
        T::F<A>: Debug + PartialEq,
    {
        eqr(
            "identity: pure id <*> v = v",
            T::seq(T::pure(|a| a), fa0()),
            fa0(),
        )
    }

    pub fn seq_respects_composition<
        T: Applicative,
        A,
        B,
        C,
        F: Fn(B) -> C,
        G: Fn(A) -> B,
        FF0: Fn() -> T::F<F>,
        FG0: Fn() -> T::F<G>,
        FA0: Fn() -> T::F<A>,
    >(
        ff0: FF0,
        fg0: FG0,
        fa0: FA0,
    ) -> R
    where
        T::F<C>: Debug + PartialEq,
    {
        eqr(
            "composition: pure (.) <*> u <*> v <*> w = u <*> (v <*> w)",
            T::seq(
                T::seq(
                    T::seq(T::pure(|f: F| |g: G| move |a| f(g(a))), ff0()),
                    fg0(),
                ),
                fa0(),
            ),
            T::seq(ff0(), T::seq(fg0(), fa0())),
        )
    }

    pub fn seq_is_homomorphic<T: Applicative, A, B, A0: Fn() -> A, F: Copy + Fn(A) -> B>(
        f: F,
        a0: A0,
    ) -> R
    where
        T::F<B>: Debug + PartialEq,
    {
        eqr(
            "homomorphism: pure f <*> pure x = pure (f x)",
            T::seq(T::pure(f), T::pure(a0())),
            T::pure(f(a0())),
        )
    }

    pub fn seq_respects_interchange<
        T: Applicative,
        A,
        B,
        F: Fn(A) -> B,
        A0: Fn() -> A,
        FF0: Fn() -> T::F<F>,
    >(
        ff0: FF0,
        a0: A0,
    ) -> R
    where
        T::F<B>: Debug + PartialEq,
    {
        eqr(
            "interchange: u <*> pure y = pure ($ y) <*> u",
            T::seq(ff0(), T::pure(a0())),
            T::seq(T::pure(|f: F| f(a0())), ff0()),
        )
    }

    pub fn seq_can_be_expressed_via_la2<
        T: Applicative,
        A,
        B,
        F: Fn(A) -> B,
        FA0: Fn() -> T::F<A>,
        FF0: Fn() -> T::F<F>,
    >(
        ff0: FF0,
        fa0: FA0,
    ) -> R
    where
        T::F<B>: Debug + PartialEq,
    {
        eqr(
            "seq via la2: (<*>) = liftA2 id",
            T::seq(ff0(), fa0()),
            T::la2(|f, a| f(a), ff0(), fa0()),
        )
    }

    pub fn fmap_can_be_expressed_via_seq<
        T: Applicative,
        A,
        B,
        F: Copy + Fn(A) -> B,
        FA0: Fn() -> T::F<A>,
    >(
        f: F,
        fa0: FA0,
    ) -> R
    where
        T::F<B>: Debug + PartialEq,
    {
        eqr(
            "fmap via seq: fmap f x = pure f <*> x",
            T::fmap(f, fa0()),
            T::seq(T::pure(f), fa0()),
        )
    }

    pub fn discard_can_be_expressed_via_seq_or_la2<
        T: Applicative,
        A,
        B,
        FA0: Fn() -> T::F<A>,
        FB0: Fn() -> T::F<B>,
    >(
        fa0: FA0,
        fb0: FB0,
    ) -> R
    where
        T::F<B>: Debug + PartialEq,
    {
        eqr(
            "discard via seq: u *> v = (id <$ u) <*> v",
            T::discard_first(fa0(), fb0()),
            T::seq(T::replace(fa0(), |b| b), fb0()),
        ) + eqr(
            "discard via la2: u <* v = liftA2 const u v",
            T::discard_second(fb0(), fa0()),
            T::la2(|b, _| b, fb0(), fa0()),
        )
    }

    pub fn bind_respects_left_identity<T: Monad, A, B, F: Copy + Fn(A) -> T::F<B>, A0: Fn() -> A>(
        f: F,
        a0: A0,
    ) -> R
    where
        T::F<B>: Debug + PartialEq,
    {
        eqr(
            "left identity: pure a >>= k = k a",
            T::bind(T::pure(a0()), f),
            f(a0()),
        )
    }

    pub fn bind_respects_right_identity<T: Monad, A, FA0: Fn() -> T::F<A>>(fa0: FA0) -> R
    where
        T::F<A>: Debug + PartialEq,
    {
        eqr(
            "right identity: m >>= bind = m",
            T::bind(fa0(), T::pure),
            fa0(),
        )
    }

    pub fn bind_is_associative<
        T: Monad,
        A,
        B,
        C,
        F: Copy + Fn(B) -> T::F<C>,
        G: Copy + Fn(A) -> T::F<B>,
        FA0: Fn() -> T::F<A>,
    >(
        f: F,
        g: G,
        fa0: FA0,
    ) -> R
    where
        T::F<C>: Debug + PartialEq,
    {
        eqr(
            r"associativity: m >>= (\x -> k x >>= h) = (m >>= k) >>= h",
            T::bind(fa0(), |a| T::bind(g(a), f)),
            T::bind(T::bind(fa0(), g), f),
        )
    }

    pub fn seq_can_be_expressed_via_bind<
        T: Monad,
        A,
        B,
        F: Fn(A) -> B,
        FA0: Fn() -> T::F<A>,
        FF0: Fn() -> T::F<F>,
    >(
        ff0: FF0,
        fa0: FA0,
    ) -> R
    where
        T::F<B>: Debug + PartialEq,
    {
        eqr(
            r"seq via bind: m1 <*> m2 = m1 >>= (\x1 -> m2 >>= (\x2 -> pure (x1 x2)))",
            T::seq(ff0(), fa0()),
            T::bind(ff0(), |f| T::bind(fa0(), |a| T::pure(f(a)))),
        )
    }

    pub fn fmap_can_be_expressed_via_bind<
        T: Monad,
        A,
        B,
        F: Copy + Fn(A) -> B,
        FA0: Fn() -> T::F<A>,
    >(
        f: F,
        fa0: FA0,
    ) -> R
    where
        T::F<B>: Debug + PartialEq,
    {
        eqr(
            "fmap via bind: fmap f xs  =  xs >>= return . f",
            T::fmap(f, fa0()),
            T::bind(fa0(), |a| T::pure(f(a))),
        )
    }
}

#[cfg(test)]
pub mod test_suite {
    use std::fmt::Debug;

    use super::tests::*;
    use super::*;

    pub trait FunctorTestSuite: WeakFunctor {
        fn sample<A>() -> Vec<Box<dyn Fn(A) -> Self::F<A>>>;
    }

    pub fn functor_follows_laws<T: Functor + FunctorTestSuite>() -> R
    where
        T::F<i32>: Debug + PartialEq,
    {
        let mut res = R::default();
        for pa in T::sample::<_>() {
            res += fmap_respects_identity::<T, _, _>(|| pa(2));
        }
        for pa in T::sample::<_>() {
            res += fmap_respects_composition::<T, _, _, _, _, _, _>(|x| x + 5, |x| x + 3, || pa(2));
        }
        res
    }

    pub fn applicative_follows_laws<T: Applicative + FunctorTestSuite>() -> R
    where
        T::F<i32>: Debug + PartialEq,
    {
        let mut res = functor_follows_laws::<T>();
        for pa in T::sample::<_>() {
            res += seq_respects_identity::<T, _, _>(|| pa(2));
        }
        for pa in T::sample::<_>() {
            for pg in T::sample::<_>() {
                for pf in T::sample::<_>() {
                    res += seq_respects_composition::<T, _, _, _, _, _, _, _, _>(
                        || pf(|x| x + 5),
                        || pg(|x| x + 3),
                        || pa(2),
                    );
                }
            }
        }
        res += seq_is_homomorphic::<T, _, _, _, _>(|x| x + 3, || 2);
        for pf in T::sample::<_>() {
            res += seq_respects_interchange::<T, _, _, _, _, _>(|| pf(|x| x + 3), || 2);
        }
        for pa in T::sample::<_>() {
            for pf in T::sample::<_>() {
                res += seq_can_be_expressed_via_la2::<T, _, _, _, _, _>(|| pf(|x| x + 3), || pa(2));
            }
        }
        for pa in T::sample::<_>() {
            res += fmap_can_be_expressed_via_seq::<T, _, _, _, _>(|x| x + 3, || pa(2));
        }
        for pa in T::sample::<_>() {
            for pb in T::sample::<_>() {
                res += discard_can_be_expressed_via_seq_or_la2::<T, _, _, _, _>(|| pa(2), || pb(2));
            }
        }
        res
    }

    pub fn monad_follows_laws<T: Monad + FunctorTestSuite>() -> R
    where
        T::F<i32>: Debug + PartialEq,
    {
        let mut res = applicative_follows_laws::<T>();
        for pa in T::sample::<_>() {
            res += bind_respects_left_identity::<T, _, _, _, _>(|x| pa(x + 3), || 2);
        }
        for pa in T::sample::<_>() {
            res += bind_respects_right_identity::<T, _, _>(|| pa(2));
        }
        for pa in T::sample::<_>() {
            for pg in T::sample::<_>() {
                for pf in T::sample::<_>() {
                    res += bind_is_associative::<T, _, _, _, _, _, _>(
                        |x| pf(x + 5),
                        |x| pg(x + 3),
                        || pa(2),
                    );
                }
            }
        }
        for pa in T::sample::<_>() {
            for pf in T::sample::<_>() {
                res += seq_can_be_expressed_via_bind::<T, _, _, _, _, _>(
                    || pf(|x| x + 3),
                    || pa(2),
                );
            }
        }
        for pa in T::sample::<_>() {
            res += fmap_can_be_expressed_via_bind::<T, _, _, _, _>(|x| x + 3, || pa(2));
        }
        res
    }
}
