use crate::func::*;

pub struct OptionClass;

impl WeakFunctor for OptionClass {
    type F<A> = Option<A>;
}

impl Functor for OptionClass {
    fn fmap<A, B, F: FnOnce(A) -> B>(f: F, fa: Self::F<A>) -> Self::F<B> {
        match fa {
            Some(a) => Some(f(a)),
            None => None,
        }
    }

    fn replace<A, B>(fa: Self::F<A>, b: B) -> Self::F<B> {
        match fa {
            Some(_) => Some(b),
            None => None,
        }
    }
}

impl ApplicativeSeq for OptionClass {
    fn seq<A, B, F: FnOnce(A) -> B>(ff: Self::F<F>, fa: Self::F<A>) -> Self::F<B> {
        match (ff, fa) {
            (Some(f), Some(a)) => Some(f(a)),
            _ => None,
        }
    }
}

impl ApplicativeLA2 for OptionClass {
    fn la2<A, B, C, F: FnOnce(A, B) -> C>(f: F, fa: Self::F<A>, fb: Self::F<B>) -> Self::F<C> {
        match (fa, fb) {
            (Some(a), Some(b)) => Some(f(a, b)),
            _ => None,
        }
    }
}

impl Applicative for OptionClass {
    fn pure<A>(a: A) -> Self::F<A> {
        Some(a)
    }

    fn discard_first<A, B>(fa: Self::F<A>, fb: Self::F<B>) -> Self::F<B> {
        match fa {
            Some(_) => fb,
            None => None,
        }
    }

    fn discard_second<A, B>(fa: Self::F<A>, fb: Self::F<B>) -> Self::F<A> {
        match fb {
            Some(_) => fa,
            None => None,
        }
    }
}

impl Monad for OptionClass {
    fn bind<A, B, F: FnOnce(A) -> Self::F<B>>(fa: Self::F<A>, f: F) -> Self::F<B> {
        match fa {
            Some(a) => f(a),
            None => None,
        }
    }

    fn join<A>(ffa: Self::F<Self::F<A>>) -> Self::F<A> {
        match ffa {
            Some(Some(a)) => Some(a),
            _ => None,
        }
    }
}

#[cfg(test)]
mod option_tests {
    use super::test_suite;

    use super::{Functor, OptionClass as T};

    impl test_suite::FunctorTestSuite for T {
        fn sample<A>() -> Vec<Box<dyn Fn(A) -> Self::F<A>>> {
            vec![Box::new(|_| None), Box::new(|a| Some(a))]
        }
    }

    #[test]
    fn fmap_f_none_is_none() {
        assert_eq!(T::fmap(|_: ()| (), None), None);
    }

    #[test]
    fn fmap_f_some_a_is_some_f_a() {
        assert_eq!(T::fmap(|x| x * x, Some(2)), Some(4));
    }

    #[test]
    fn replace_none_b_is_none() {
        assert_eq!(T::replace(None::<i32>, 1), None);
        assert_eq!(T::void(None::<i32>), None);
    }

    #[test]
    fn replace_some_a_b_is_some_b() {
        assert_eq!(T::replace(Some(1), 2), Some(2));
        assert_eq!(T::void(Some(1)), Some(()));
    }

    #[test]
    fn monad_follows_laws() {
        test_suite::monad_follows_laws::<T>().unwrap();
    }
}
